#include "tiny-06.hpp"

#include "catch.hpp"


// Nejméně záruk, neportabilní -- musí se kontrolovat pouze zaručené intervaly
TEST_CASE("Check random generator", "[random][all]") {
    auto rand_gen = numbers_generator::create_by_kind(generator_kind::random, 1234);
    CHECK(rand_gen->bools_generated() == 0);
    CHECK(rand_gen->doubles_generated() == 0);
    CHECK(rand_gen->ints_generated() == 0);
    SECTION("Generate bools until we get 10 falses") {
        size_t num_false = 0;
        while (num_false < 10) {
            auto res = rand_gen->random_bool();
            if (!res) { ++num_false; }
        }
        REQUIRE(rand_gen->bools_generated() >= 10);
        REQUIRE(rand_gen->falses_generated() == 10);
        REQUIRE(rand_gen->trues_generated() == rand_gen->bools_generated() - 10);
    }
    SECTION("Generate doubles until we get 10 small ones") {
        size_t num_small = 0;
        size_t num_big = 0;
        while (num_small < 10) {
            auto res = rand_gen->random_double();
            if (res < 0.1) { ++num_small; }
            else { ++num_big; }
        }
        REQUIRE(rand_gen->doubles_generated() == num_small + num_big);
        REQUIRE(rand_gen->min_generated_double() <  0.1);
        REQUIRE(rand_gen->min_generated_double() >= 0.0);
        REQUIRE(rand_gen->max_generated_double() >= 0.1);
        REQUIRE(rand_gen->max_generated_double() <  1.0);
    }
    SECTION("Generate ints from significantly different ranges") {
        for (int i = 0; i < 16; i += 4) {
            auto res = rand_gen->random_int(-200 + i * 100, -100 + i * 100);
            CHECK(res >= -200 + i * 100);
            CHECK(res <= -100 + i * 100);
        }
        CHECK(rand_gen->ints_generated() == 4);
        REQUIRE(rand_gen->min_generated_int() >= -200);
        REQUIRE(rand_gen->min_generated_int() <= -100);
        REQUIRE(rand_gen->max_generated_int() >= 1000);
        REQUIRE(rand_gen->max_generated_int() <= 1100);
    }
}

TEST_CASE("mt generator", "[twister][all]") {
    auto mt_gen = numbers_generator::create_by_kind(generator_kind::mersenne, 1234);
    CHECK(mt_gen->bools_generated() == 0);
    CHECK(mt_gen->doubles_generated() == 0);
    CHECK(mt_gen->ints_generated() == 0);
    SECTION("Run coinflips for a while") {
        auto coinflips = mt_gen->random_int(40, 80);
        CHECK(coinflips == 42);
        for (int i = 0; i < coinflips; ++i) {
            static_cast<void>(mt_gen->random_bool());
        }
        CHECK(mt_gen->bools_generated() == 42);
        REQUIRE(mt_gen->trues_generated() == 18);
        REQUIRE(mt_gen->falses_generated() == 24);
    }
    SECTION("Generate some doubles") {
        auto num_doubles = mt_gen->random_int(50, 100);
        CHECK(num_doubles == 53);
        for (int i = 0; i < num_doubles; ++i) {
            static_cast<void>(mt_gen->random_double());
        }
        CHECK(mt_gen->doubles_generated() == 53);
        REQUIRE(mt_gen->min_generated_double() == Approx(0.0153917951));
        REQUIRE(mt_gen->max_generated_double() == Approx(0.987487716));
    }
    SECTION("Generate some ints") {
        auto num_ints = mt_gen->random_int(60, 120);
        CHECK(num_ints == 64);
        for (int i = 0; i < num_ints; ++i) {
            static_cast<void>(mt_gen->random_int(i, 200));
        }
        CHECK(mt_gen->ints_generated() == 65);
        REQUIRE(mt_gen->min_generated_int() ==  10);
        REQUIRE(mt_gen->max_generated_int() == 200);
    }
}

TEST_CASE("minstd generator", "[minstd][all]") {
    auto minstd_gen = numbers_generator::create_by_kind(generator_kind::minstd, 1234);
    CHECK(minstd_gen->bools_generated() == 0);
    CHECK(minstd_gen->doubles_generated() == 0);
    CHECK(minstd_gen->ints_generated() == 0);
    SECTION("Run coinflips for a while") {
        auto coinflips = minstd_gen->random_int(40, 80);
        CHECK(coinflips == 41);
        for (int i = 0; i < coinflips; ++i) {
            static_cast<void>(minstd_gen->random_bool());
        }
        CHECK(minstd_gen->bools_generated() == 41);
        REQUIRE(minstd_gen->trues_generated() == 23);
        REQUIRE(minstd_gen->falses_generated() == 18);
    }
    SECTION("Generate some doubles") {
        auto num_doubles = minstd_gen->random_int(50, 100);
        CHECK(num_doubles == 51);
        for (int i = 0; i < num_doubles; ++i) {
            static_cast<void>(minstd_gen->random_double());
        }
        CHECK(minstd_gen->doubles_generated() == 51);
        REQUIRE(minstd_gen->min_generated_double() == Approx(0.0052501038));
        REQUIRE(minstd_gen->max_generated_double() == Approx(0.9801495811));
    }
    SECTION("Generate some ints") {
        auto num_ints = minstd_gen->random_int(60, 120);
        CHECK(num_ints == 61);
        for (int i = 0; i < num_ints; ++i) {
            static_cast<void>(minstd_gen->random_int(i, 200));
        }
        CHECK(minstd_gen->ints_generated() == 62);
        REQUIRE(minstd_gen->min_generated_int() ==  12);
        REQUIRE(minstd_gen->max_generated_int() == 197);
    }
}
